#include "string.h"
#include "printf.h"
#include "failure.h"
#include "page.h"
#include "multiboot.h"
#include "system.h"
#include "msr.h"
#include "elf.h"
#include "vm.h"
#include "svm.h"
#include "alloc.h"
#include "vmexit.h"


#include "vmm.h" /* [DEBUG] */


static struct vmcb *
alloc_vmcb ( void )
{
	struct vmcb *vmcb;

	const unsigned long pfn = alloc_pages ( 1, 1 );
	vmcb = ( struct vmcb * ) VIRT ( pfn << PAGE_SHIFT );
	memset ( ( char * ) vmcb, 0, sizeof ( struct vmcb ) );

	return vmcb;
}

static unsigned long 
create_intercept_table ( unsigned long size )
{
	const unsigned long pfn = alloc_pages ( size >> PAGE_SHIFT, 1 );
	void *p = ( void * ) VIRT ( pfn << PAGE_SHIFT );

	/* vol. 2, p. 445 */
	memset ( p, 0xff, size );

	return pfn << PAGE_SHIFT; 
}


static unsigned long
alloc_vm_pmem ( unsigned long size )
{
	const unsigned long pfn = alloc_pages ( size >> PAGE_SHIFT, 1 );
	return ( unsigned long ) VIRT ( pfn << PAGE_SHIFT );
}

static struct multiboot_info *
init_vm_mbi ( unsigned long vm_pmem_start )
{
	enum { INSTALL_PADDR = 0x2d0e0UL }; /* < 1 MB */
	struct multiboot_info *mbi = ( struct multiboot_info *) ( vm_pmem_start + INSTALL_PADDR );

	/* [TODO] */

	printf ( "Multiboot information initialized.\n" );

	return mbi;
}

static int
is_reserved_pmem ( unsigned long pfn )
{
#if 0
	return ( pfn < ( 1 << ( 20 - PAGE_SHIFT ) ) );
#else
	/* [DEBUG] */
	return 0;
#endif
}

static void
mmap_pages ( unsigned long cr3, unsigned long vm_pmem_start, unsigned long vm_pmem_size )
{
	const unsigned long vm_pmem_pfn = PHYS ( vm_pmem_start ) >> PAGE_SHIFT;
	int i;

	for ( i = 0; i < PFN_UP ( vm_pmem_size ); i++ ) {
		const unsigned long pml4     = ( unsigned long ) VIRT ( cr3 );
		const unsigned long vm_paddr = i << PAGE_SHIFT;
		const unsigned long pm_pfn   = i + ( is_reserved_pmem ( i ) ? 0 : vm_pmem_pfn ); /* [NAIVE] for VGA */
		const unsigned long pm_paddr = pm_pfn << PAGE_SHIFT;
		mmap ( pml4, vm_paddr, pm_paddr );
	}
}

/* ۥޥʪ¥ޥʪбŤڡơ֥
 * ʪɥ쥹֤*/
static unsigned long 
create_shadow_table ( unsigned long vm_pmem_start, unsigned long vm_pmem_size )
{
	const unsigned long cr3 = pml4_table_create ( );
	mmap_pages ( cr3, vm_pmem_start, vm_pmem_size );
	printf ( "Shadow table created (vaddr=%x, size=%x).\n", vm_pmem_start, vm_pmem_size );
	return cr3;
}

static unsigned long 
create_initial_pg_table ( void )
{
	const unsigned long cr3 = pml4_table_create ( );
	mmap_pages ( cr3, ( unsigned long ) VIRT ( 0 ), 1UL << 30 );
	printf ( "Initial table created.\n" );
	return cr3;
}


/* [DEBUG] */
#if 0

static void	
set_control_area ( struct vmcb *vmcb )
{
	/* Enable nested paging (See AMD64 manual Vol. 3, p. 488) */
//	vmcb->np_enable = 1; 

	/* Flush TLB on VMRUN (all entries, all ASIDs) */
        vmcb->tlb_control = 1;

	/* To be added in RDTSC and RDTSCP */
	vmcb->tsc_offset = 0; 
	
	/* Guest address space identifier (ASID) */
	vmcb->guest_asid = 1;

	vmcb->general2_intercepts = INTRCPT_VMRUN;

	/* [REF] vol.2, p. 454 */
	vmcb->iopm_base_pa  = create_intercept_table ( 12 << 10 /* 12 Kbytes */ );
	vmcb->msrpm_base_pa = create_intercept_table ( 8 << 10 /* 8 Kbytes */ );
}

static void
set_descriptors ( struct vmcb *vmcb )
{
	int i;

	/* Generic Segment Descriptor - Legacy Mode 
	 * (AMD64 manual Vol 2, pp. 98-104) */

	/* setup the segment registers and all their hidden states */
	vmcb->cs.sel   = 0xF000; 
	vmcb->cs.attrs.bytes = 0x089b; // 1000 1001 1011
	vmcb->cs.limit = 0xffff;
	vmcb->cs.base  = 0x000F0000;

	// type = 1011 : Data-segment, R(eadable), A(ccessed)
	// S, 
	// P(resent), 
	// G(ranularity)

	struct seg_selector *segregs [] = { &vmcb->ss, &vmcb->ds, &vmcb->es, &vmcb->fs, &vmcb->gs, NULL };
	for ( i = 0; segregs [ i ] != NULL; i++ ) {
		struct seg_selector *x = segregs [ i ];
		x->sel   = 0x00;
		x->attrs.bytes = 0x0893; // 1000 1001 0011 
		x->limit = 0xffff;
		x->base  = 0x00;

		// type = 0011 : Data-segment, W(ritable), A(ccessed)
		// S, 
		// P(resent), 
		// G(ranularity)
	}
	
	struct seg_selector *dsegs [] = { &vmcb->ldtr, &vmcb->gdtr, &vmcb->tr, NULL };
	for ( i = 0; dsegs [ i ] != NULL; i++ ) {
		struct seg_selector *x = dsegs [ i ];
		x->sel   = 0x00;
		x->attrs.bytes = 0x0000;
		x->limit = 0x0;
		x->base  = 0x00;
	}

	vmcb->idtr.sel   = 0x00;
	vmcb->idtr.attrs.bytes = 0x0000;
	vmcb->idtr.limit = 0x3ff;
	vmcb->idtr.base  = 0x00;
}

static void
set_state_save_area ( struct vmcb *vmcb )
{
	set_descriptors ( vmcb );

	vmcb->cpl = 0; /* must be equal to SS.DPL  */

	vmcb->rax = 0;
	vmcb->rsp = 0;

	vmcb->rflags = 2;

//	vmcb->cr3 = pagetable_get_paddr ( svm_paging_enabled(v) ? v->arch.guest_table : ov->domain->arch.phys_table );
	
	/* (Paged real mode) VMRUN can load a guest value of CR0 with PE = 0 and PG = 1,
	 * a combination that is otherwise illegal (See AMD64 manual vol. 2, p. 445, p. 467). */
//	vmcb->cr0 = X86_CR0_ET | X86_CR0_PG;
	vmcb->cr0 = X86_CR0_ET; 

//	vmcb->cr4 = X86_CR4_PAE;

	vmcb->efer = EFER_SVME;
}

/* Create a new virtual machine with a given OS image.
 *  [input]
 *    - guest_image_start : 
 *    - guest_image_size : 
 *    - guest_pmem_size :
 *  [output]
 *    - vm : 
 */
void
vm_create ( struct vm *vm, unsigned long guest_image_start, unsigned long guest_image_size, unsigned long vm_pmem_size )
{
	/* Allocate a new page for storing VMCB.  */
	vm->vmcb = alloc_vmcb ( );

	/* Allocate new pages for physical memory of the guest OS.  */
	const unsigned long vm_pmem_start = alloc_vm_pmem ( vm_pmem_size );

	/* Copy the OS image to the specified region by interpreting the ELF format.  */
	vm->vmcb->rip = load_elf_image ( guest_image_start, guest_image_size, vm_pmem_start );

	/* Setup multiboot info.  */
	vm->mbi = init_vm_mbi ( vm_pmem_start );

	/* Set Host-level CR3 to use for nested paging.  */
	vm->h_cr3 = create_shadow_table ( vm_pmem_start, vm_pmem_size );
//	vm->vmcb->h_cr3 = vm->h_cr3;

	set_control_area ( vm->vmcb );
	set_state_save_area ( vm->vmcb );

	printf ( "New virtual machine created.\n" ); 
}



#else

// #define ENABLE_NP_PAGING


void 
print_something ( )
{
	printf ( "******* VM ******\n" );
}

// vmcb->g_pat ?? (guest pat) p. 230

/* Host OS Ʊ Guest OS  */
void
vm_create ( struct vm *vm, unsigned long guest_image_start, unsigned long guest_image_size, unsigned long vm_pmem_size )
{
	struct vmcb *vmcb;

	vmcb = alloc_vmcb ( );
	vm->vmcb = vmcb;

#ifdef ENABLE_NP_PAGING
	const unsigned long vm_pmem_start = alloc_vm_pmem ( vm_pmem_size );
	memmove ( ( void * )vm_pmem_start, VIRT ( 0 ), vm_pmem_size ); // copy guest 
	vm->h_cr3 = create_shadow_table ( vm_pmem_start, vm_pmem_size );

	printf ( "Nested paging is enabled.\n" );
	vmcb->np_enable = 1;  /* ͭˤ VMEXIT_INVAILD ȯ */
	vmcb->h_cr3 = vm->h_cr3; 
#else
	/* #VMEXIT: page fault */
	printf ( "Nested paging is disabled.\n" );
#endif

	vmcb->guest_asid  = 1;
        vmcb->tlb_control = 1;	
	vmcb->general2_intercepts = INTRCPT_VMRUN;
	vmcb->iopm_base_pa  = create_intercept_table ( 12 << 10 /* 12 Kbytes */ );
	vmcb->msrpm_base_pa = create_intercept_table ( 8 << 10  /* 8 Kbytes */ );

	vm->vmcb->rip = ( unsigned long )( &print_something );
	printf ( "vmcb->rip = %x\n", vm->vmcb->rip );

#if 1

	unsigned long rsp;
	asm ( "movq %%rsp, %0" : "=r" ( rsp ) );
	vm->vmcb->rsp = rsp;
	printf ( "vmcb->rsp = %x\n", vm->vmcb->rsp );

	/* setup the segment registers and all their hidden states */
	vmcb->cs.sel   = VMM_CS64;
//	vmcb->cs.attrs.bytes = 0x089b; // 1000 1001 1011
	vmcb->cs.attrs.fields.type = 0xa;
	vmcb->cs.attrs.fields.l = 1;
	vmcb->cs.attrs.fields.g = 1;
	vmcb->cs.limit = 0xffff;
	vmcb->cs.base  = 0;

	print_binary ( &vmcb->cs.attrs.bytes, 2 );

	// type = 1011 : Code-segment, R(eadable), A(ccessed)
	// S, 
	// P(resent), 
	// G(ranularity)

	struct seg_selector *segregs [] = { &vmcb->ss, &vmcb->ds, &vmcb->es, &vmcb->fs, &vmcb->gs, NULL };
	int i;
	for ( i = 0; segregs [ i ] != NULL; i++ ) {
		struct seg_selector *x = segregs [ i ];
		x->sel   = VMM_DS64;
		x->attrs.bytes = 0x0893; // 1000 1001 0011 
		x->limit = 0xffff;
		x->base  = 0;

		// type = 0011 : Data-segment, W(ritable), A(ccessed)
		// S, 
		// P(resent), 
		// G(ranularity)
	}

	u64 dtr;
	__asm__ __volatile__ ("sgdt %0":"=m" ( dtr ) );
	printf  ( "\n" "sgdt: limit=%x, base=%x\n", dtr & 0xff, dtr >> 16 );
	vmcb->gdtr.sel   = 0;
	vmcb->gdtr.attrs.bytes = 0;
	vmcb->gdtr.limit = dtr & 0xff;
	vmcb->gdtr.base  = dtr >> 16;
	
	unsigned long rflags;
	__asm__ __volatile__( "pushfq; popq %0" :"=m" ( rflags ) );
	vmcb->rflags = rflags & 2;

	vmcb->cr0 = read_cr0 ();
	vmcb->cr3 = read_cr3 ();
	vmcb->cr4 = read_cr4 ();

	u32 eax, edx;
	rdmsr ( MSR_EFER, eax, edx );
	// vmcb->efer = EFER_SVME;
	vmcb->efer = ( ( ( (u64) edx) << 32 ) | eax );
#endif

	printf ( "New virtual machine created.\n" ); 
}

#endif

/******************************************************/

void
vm_boot ( struct vm *vm )
{
	printf ( "Booting guest operating system.\n" ); 

	vmcb_check_consistency ( vm->vmcb );

	/* Switch from VMM to guest OS */
	u64 p_vmcb = PHYS ( vm->vmcb );
	svm_launch ( p_vmcb );

	/* VMEXIT ˤ return */

	print_vmexit_exitcode ( vm->vmcb->exitcode );
	printf ( "vmcb->cr2=%x, vmcb->rip=%x\n", vm->vmcb->cr2, vm->vmcb->rip );
	printf ( "vmcb->rsp=%x\n", vm->vmcb->rsp );
        printf ( "fault addr=%x, error_code=%x\n", vm->vmcb->exitinfo2, vm->vmcb->exitinfo1 );

	printf ( "\n" "sel=%x, limit=%x, base=%x\n",
		 vm->vmcb->cs.sel, vm->vmcb->cs.limit, vm->vmcb->cs.base );

	print_binary ( &vm->vmcb->cs.attrs.bytes, 2 );
}



